import shutil
import subprocess
from pathlib import Path
from subprocess import CalledProcessError
from typing import Optional, Tuple, Union

from rich.console import Console

console = Console()

import click
from rich import print

DEFAULT_OUT = "_whaler"
HTML_DIR = "html"
DU_FILENAME = "du.txt"
DU_ARGS = ("-a", "-k")
UI_FILENAME = "html.tgz"


@click.command()
@click.option("--image", help="docker image", default=None)
@click.option("--port", help="port to serve the outputs on", default=8000)
@click.option("--out", help="output path", default=DEFAULT_OUT)
@click.argument("directory", default=".")
@click.option(
    "--server/--no-server",
    default=True,
    help="Run web server on output to immediately view",
)
def run(directory: str, out: str, image: Optional[str], server: bool, port: int):
    """"""
    try:
        du_out = du(Path(directory), image)

        out_path = Path(out)
        html_path = out_path / HTML_DIR
        html_path.mkdir(exist_ok=True, parents=True)
        (html_path / DU_FILENAME).write_text(du_out)

        ui = Path(__file__).parent / "static" / UI_FILENAME

        shutil.unpack_archive(ui, extract_dir=out_path)

        if server:
            server_cmd = (
                "python3",
                "-m",
                "http.server",
                str(port),
                f"--directory={html_path}",
            )
            print(
                f"[bold green]Done. Serving output at http://localhost:{port} (ctrl+c to exit)"
            )
            shell(server_cmd)
    except CalledProcessError as err:
        msg = f"""
[bold red]subprocess failed with status {err.returncode}.
stderr:\n\n{(err.stderr or b'-').decode()}
stdout (truncated):\n\n{(err.stdout or b'-').decode()[:100]}
"""
        print(msg)

        exit(2)


def du(directory: Path, image: Optional[str]) -> str:

    if image:
        cmd = ("docker", "run", "--rm", "--entrypoint=du")
        if str(directory) != ".":
            cmd += (f"--workdir={directory}",)
        cmd += (image, *DU_ARGS)
    else:
        cmd = ("bash", "-c", f"cd {directory} && du {' '.join(DU_ARGS)}")

    try:
        return shell(cmd)
    except CalledProcessError as err:
        if len(err.output) == 0:
            raise err

        print(
            f"Ignoring what seems to be non-fatal error(s):\n{(err.stderr or b'-').decode()}"
        )
        return err.output.decode()


def shell(cmd: Union[str, Tuple[str, ...]]):
    with console.status(""):

        if isinstance(cmd, str):
            print(f"Running {cmd}")
            return subprocess.check_output(
                cmd, shell=True, stderr=subprocess.PIPE
            ).decode()
        else:
            print(f"Running {' '.join(cmd)}")
            return subprocess.check_output(cmd, stderr=subprocess.PIPE).decode()


class ShellError(BaseException):
    pass


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['whaler']

package_data = \
{'': ['*'], 'whaler': ['static/*']}

install_requires = \
['click>=7.1.2,<8.0.0',
 'humanfriendly>=9.1,<10.0',
 'pathlib>=1.0.1,<2.0.0',
 'rich>=9.5.1,<10.0.0']

entry_points = \
{'console_scripts': ['whaler = whaler.cli:run']}

setup_kwargs = {
    'name': 'whaler',
    'version': '0.1.2',
    'description': '',
    'long_description': "# whaler\n\n[![PyPI versions](https://img.shields.io/pypi/pyversions/whaler?logo=python&logoColor=white)](https://pypi.org/project/whaler)\n[![PyPI versions](https://img.shields.io/pypi/v/whaler?logo=python&logoColor=white)](https://pypi.org/project/whaler)\n\n**What?** A command-line tool for visually investigating the disk usage of docker images\n\n**Why?** Large images are slow to move and expensive to store. They cost developer productivity by lengthening devops tasks and often contain unnecessary data\n\n**Who is this for?** Primarily for engineers working with images containing Python packages.\n\n## User Stories\n\nThis tool should allow you to answer questions such as:\n1. Which file types are occupying the most disk space?\n2. Which are my largest Python packages?\n3. What are my unknown causes of high disk usage?\n\n## Quick start\n\n```bash\npip install whaler\n```\n\n### Run against a local directory\n```\n➜ whaler .venv\nRunning bash -c cd .venv && du -a -k\nDone. Serving output at http://localhost:8000 (ctrl+c to exit)\nRunning python3 -m http.server 8000 --directory=_whaler/html\n```\n\n### Run against a docker image\n\nThe tool will pull the image first if it is not present.\n```\nwhaler --image='hl:latest' /\nRunning docker run --rm --entrypoint=du --workdir=/ hl:latest -a -k\nIgnoring what seems to be non-fatal error(s):\ndu: cannot access './proc/1/task/1/fd/4': No such file or directory\ndu: cannot access './proc/1/task/1/fdinfo/4': No such file or directory\ndu: cannot access './proc/1/fd/3': No such file or directory\ndu: cannot access './proc/1/fdinfo/3': No such file or directory\n\n\nDone. Serving output at http://localhost:8000 (ctrl+c to exit)\nRunning python3 -m http.server 8000 --directory=_whaler/html\n```\n\n![HTML Report](docs/screen.png)\n\nPlay with a [hosted demo](https://treebeardtech.github.io/whaler/)\n\n## Limitations\n\n1. Platform: whaler uses `du` to gather disk usage data. It must be present in your docker image\n2. Scale: I have tested the web UI with up to 500,000 file system nodes with `du` output of up to ~100MB.\n\n## Alternatives/Complements to this tool:\n\n1. Whaler can tell you what is taking up space in the final layer of your Docker image, but you may have intermediate layers which are contributing to the image size. For diving through the layers, use [dive](https://github.com/wagoodman/dive)\n    * **Related**: read up on [multi-stage builds](https://docs.docker.com/develop/develop-images/multistage-build/) to understand how to mitigate the problem of intermediate layers bloating your image.\n1. For investigating disk usage in non-docker directories, [Disk Inventory X](http://www.derlien.com/) is a great tool on OS X which I have based whaler on.\n\n## Developing\n\nSee `.github/workflows/test.yml` for the development platform and setup.\n\nFor UI, see [whaler-ui](https://github.com/treebeardtech/whaler-ui)",
    'author': 'alex-treebeard',
    'author_email': 'alex@treebeard.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/treebeardtech/whaler',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)
